#!/usr/bin/env python3
# pylint: disable=missing-docstring,invalid-name
# pylint: disable=unused-argument
# pylint: disable=unnecessary-lambda

import pathlib
import re
import reprlib
import textwrap
import unittest
from typing import Optional, List, Tuple, Any  # pylint: disable=unused-import

import numpy

import icontract._represent
import tests.error
import tests.mock


class TestReprValues(unittest.TestCase):
    def test_num(self) -> None:
        @icontract.require(lambda x: x < 5)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=100)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "x < 5: x was 100", tests.error.wo_mandatory_location(str(violation_error))
        )

    def test_str(self) -> None:
        @icontract.require(lambda x: x != "oi")
        def func(x: str) -> str:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x="oi")
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            """x != "oi": x was 'oi'""",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_bytes(self) -> None:
        @icontract.require(lambda x: x != b"oi")
        def func(x: bytes) -> bytes:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=b"oi")
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            """x != b"oi": x was b'oi'""",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_bool(self) -> None:
        @icontract.require(lambda x: x is not False)
        def func(x: bool) -> bool:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=False)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "x is not False: x was False",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_list(self) -> None:
        y = 1

        @icontract.require(lambda x: sum([1, y, x]) == 1)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                sum([1, y, x]) == 1:
                sum([1, y, x]) was 5
                x was 3
                y was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_tuple(self) -> None:
        y = 1

        @icontract.require(lambda x: sum((1, y, x)) == 1)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                sum((1, y, x)) == 1:
                sum((1, y, x)) was 5
                x was 3
                y was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_set(self) -> None:
        y = 2

        @icontract.require(lambda x: sum({1, y, x}) == 1)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                sum({1, y, x}) == 1:
                sum({1, y, x}) was 6
                x was 3
                y was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_dict(self) -> None:
        y = "someKey"

        @icontract.require(lambda x: len({y: 3, x: 8}) == 6)
        def func(x: str) -> str:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x="oi")
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                len({y: 3, x: 8}) == 6:
                len({y: 3, x: 8}) was 2
                x was 'oi'
                y was 'someKey'"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_unary_op(self) -> None:
        @icontract.require(lambda x: not -x + 10 > 3)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "not -x + 10 > 3: x was 1",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_binary_op(self) -> None:
        @icontract.require(lambda x: -x + x - x * x / x // x**x % x > 3)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "-x + x - x * x / x // x**x % x > 3: x was 1",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_binary_op_bit(self) -> None:
        @icontract.require(lambda x: ~(x << x | x & x ^ x) >> x > x)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "~(x << x | x & x ^ x) >> x > x: x was 1",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_bool_op_single(self) -> None:
        # pylint: disable=chained-comparison
        @icontract.require(lambda x: x > 3 and x < 10)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "x > 3 and x < 10: x was 1",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_bool_op_multiple(self) -> None:
        # pylint: disable=chained-comparison
        @icontract.require(lambda x: x > 3 and x < 10 and x % 2 == 0)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "x > 3 and x < 10 and x % 2 == 0: x was 1",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_compare(self) -> None:
        # pylint: disable=chained-comparison

        # Chain the compare operators in a meaningless order and semantics
        @icontract.require(
            lambda x: 0 < x < 3
            and x > 10
            and x != 7
            and x >= 10
            and x <= 11
            and x is not None
            and x in [1, 2, 3]
            and x not in [1, 2, 3]
        )
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            """\
0 < x < 3
    and x > 10
    and x != 7
    and x >= 10
    and x <= 11
    and x is not None
    and x in [1, 2, 3]
    and x not in [1, 2, 3]: x was 1""",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_call(self) -> None:
        def y() -> int:
            return 1

        @icontract.require(lambda x: x < y())
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < y():
                x was 1
                y() was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_if_exp_body(self) -> None:
        y = 5

        @icontract.require(lambda x: x < (x**2 if y == 5 else x**3))
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < (x**2 if y == 5 else x**3):
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_if_exp_orelse(self) -> None:
        y = 5

        @icontract.require(lambda x: x < (x**2 if y != 5 else x**3))
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < (x**2 if y != 5 else x**3):
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_attr(self) -> None:
        class A:
            def __init__(self) -> None:
                self.y = 3

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        a = A()

        @icontract.require(lambda x: x > a.y)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > a.y:
                a was an instance of A
                a.y was 3
                x was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_index(self) -> None:
        lst = [1, 2, 3]

        @icontract.require(lambda x: x > lst[1])
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > lst[1]:
                lst was [1, 2, 3]
                lst[1] was 2
                x was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_slice(self) -> None:
        lst = [1, 2, 3]

        @icontract.require(lambda x: x > sum(lst[1:2:1]))
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > sum(lst[1:2:1]):
                lst was [1, 2, 3]
                lst[1:2:1] was [2]
                sum(lst[1:2:1]) was 2
                x was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_ext_slice(self) -> None:
        class SomeClass:
            def __getitem__(self, item: Any) -> Any:
                return item

            def __repr__(self) -> str:
                return "<instance of SomeClass>"

        @icontract.require(lambda something: something[1, 2:3] is None)
        def func(something: SomeClass) -> None:
            pass

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            an_instance = SomeClass()
            func(something=an_instance)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            textwrap.dedent(
                """\
                something[1, 2:3] is None:
                something was <instance of SomeClass>
                something[1, 2:3] was (1, slice(2, 3, None))"""
            ),
            tests.error.wo_mandatory_location(str(violation_err)),
        )

    def test_lambda(self) -> None:
        @icontract.require(lambda x: x > (lambda y: y + 4).__call__(y=7))  # type: ignore
        def func(x: int) -> int:
            return x

        runtime_error = None  # type: Optional[RuntimeError]
        try:
            func(x=1)
        except RuntimeError as err:
            runtime_error = err

        assert runtime_error is not None
        assert runtime_error.__cause__ is not None
        assert isinstance(runtime_error.__cause__, NotImplementedError)

        not_implemented_error = runtime_error.__cause__

        self.assertEqual(
            "Re-computation of in-line lambda functions is not supported since it is quite tricky to implement and "
            "we decided to implement it only once there is a real need for it. "
            "Please make a feature request on https://github.com/Parquery/icontract",
            str(not_implemented_error),
        )


class TestGeneratorExpr(unittest.TestCase):
    def test_attr_on_element(self) -> None:
        @icontract.ensure(
            lambda result: all(single_res[1].is_absolute() for single_res in result)
        )
        def some_func() -> List[Tuple[pathlib.Path, pathlib.Path]]:
            return [(pathlib.Path("/home/file1"), pathlib.Path("home/file2"))]

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func()
        except icontract.ViolationError as err:
            violation_error = err

        # This dummy path is necessary to obtain the class name.
        dummy_path = pathlib.Path("/also/doesnt/exist")

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                all(single_res[1].is_absolute() for single_res in result):
                all(single_res[1].is_absolute() for single_res in result) was False, e.g., with
                  single_res = ({0}('/home/file1'), {0}('home/file2'))
                result was [({0}('/home/file1'), {0}('home/file2'))]"""
            ).format(dummy_path.__class__.__name__),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_multiple_for(self) -> None:
        lst = [[1, 2], [3]]

        # fmt: off
        @icontract.require(
            lambda x: all(item == x for sublst in lst for item in sublst)
        )
        # fmt: on
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)

        self.assertEqual(
            textwrap.dedent(
                """\
                all(item == x for sublst in lst for item in sublst):
                all(item == x for sublst in lst for item in sublst) was False, e.g., with
                  sublst = [1, 2]
                  item = 1
                lst was [[1, 2], [3]]
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_zip_and_multiple_for(self) -> None:
        # Taken from a solution for Advent of Code 2020 day 11.
        @icontract.ensure(
            lambda layout, result: all(
                cell == result_cell
                for row, result_row in zip(layout, result[0])
                for cell, result_cell in zip(row, result_row)
                if cell == "."
            ),
            "Floor remains floor",
        )
        def apply(layout: List[List[str]]) -> Tuple[List[List[str]], int]:
            height = len(layout)
            width = len(layout[0])

            result = [[""] * width] * height
            return result, 0

        layout = [["L", ".", "#"], [".", "#", "#"]]

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _, _ = apply(layout=layout)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)

        text = re.sub(
            r"<zip object at 0x[0-9a-fA-F]+>",
            "<zip object at some address>",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                Floor remains floor: all(
                        cell == result_cell
                        for row, result_row in zip(layout, result[0])
                        for cell, result_cell in zip(row, result_row)
                        if cell == "."
                    ):
                all(
                        cell == result_cell
                        for row, result_row in zip(layout, result[0])
                        for cell, result_cell in zip(row, result_row)
                        if cell == "."
                    ) was False, e.g., with
                  row = ['L', '.', '#']
                  result_row = ['', '', '']
                  cell = '.'
                  result_cell = ''
                layout was [['L', '.', '#'], ['.', '#', '#']]
                result was ([['', '', ''], ['', '', '']], 0)
                result[0] was [['', '', ''], ['', '', '']]
                zip(layout, result[0]) was <zip object at some address>"""
            ),
            text,
        )


class TestListComprehension(unittest.TestCase):
    def test_single(self) -> None:
        lst = [1, 2, 3]

        @icontract.require(lambda x: [item < x for item in lst if item % x == 0] == [])
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                [item < x for item in lst if item % x == 0] == []:
                [item < x for item in lst if item % x == 0] was [False]
                lst was [1, 2, 3]
                x was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_nested(self) -> None:
        lst_of_lsts = [[1, 2, 3]]

        # fmt: off
        @icontract.require(
            lambda:
            [
                [item for item in sublst if item > 0]
                for sublst in lst_of_lsts
            ] == [[]]
        )
        # fmt: on
        def func() -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                [
                        [item for item in sublst if item > 0]
                        for sublst in lst_of_lsts
                    ] == [[]]:
                [
                        [item for item in sublst if item > 0]
                        for sublst in lst_of_lsts
                    ] was [[1, 2, 3]]
                lst_of_lsts was [[1, 2, 3]]"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestSetComprehension(unittest.TestCase):
    def test_single(self) -> None:
        lst = [1, 2, 3]

        @icontract.require(
            lambda x: len({item < x for item in lst if item % x == 0}) == 0
        )
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                len({item < x for item in lst if item % x == 0}) == 0:
                len({item < x for item in lst if item % x == 0}) was 1
                lst was [1, 2, 3]
                x was 2
                {item < x for item in lst if item % x == 0} was {False}"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_nested(self) -> None:
        lst_of_lsts = [[1, 2, 3]]

        # fmt: off
        @icontract.require(
            lambda:
            {
                len({item for item in lst if item > 0})
                for lst in lst_of_lsts
            } == set()
        )
        # fmt: on
        def func() -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                {
                        len({item for item in lst if item > 0})
                        for lst in lst_of_lsts
                    } == set():
                lst_of_lsts was [[1, 2, 3]]
                set() was set()
                {
                        len({item for item in lst if item > 0})
                        for lst in lst_of_lsts
                    } was {3}"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestDictComprehension(unittest.TestCase):
    def test_single(self) -> None:
        @icontract.require(lambda x: len({i: i**2 for i in range(x)}) == 0)
        def func(x: int) -> int:
            return x

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                len({i: i**2 for i in range(x)}) == 0:
                len({i: i**2 for i in range(x)}) was 2
                range(x) was range(0, 2)
                x was 2
                {i: i**2 for i in range(x)} was {0: 0, 1: 1}"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_nested(self) -> None:
        lst_of_lsts = [[1, 2, 3]]

        # fmt: off
        @icontract.require(
            lambda:
            len({
                len(lst): {item: item for item in lst}
                for lst in lst_of_lsts
            }) == 0
        )
        # fmt: on
        def func() -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                len({
                        len(lst): {item: item for item in lst}
                        for lst in lst_of_lsts
                    }) == 0:
                len({
                        len(lst): {item: item for item in lst}
                        for lst in lst_of_lsts
                    }) was 1
                lst_of_lsts was [[1, 2, 3]]
                {
                        len(lst): {item: item for item in lst}
                        for lst in lst_of_lsts
                    } was {3: {1: 1, 2: 2, 3: 3}}"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestConditionAsText(unittest.TestCase):
    """Test decompilation of the condition."""

    def test_single_line(self) -> None:
        # fmt: off
        @icontract.require(lambda x: x > 3)
        def func(x: int) -> int:
            return x

        # fmt: on

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            "x > 3: x was 0", tests.error.wo_mandatory_location(str(violation_err))
        )

    def test_condition_on_next_line(self) -> None:
        # fmt: off
        @icontract.require(
            lambda x: x > 3)
        def func(x: int) -> int:
            return x

        # fmt: on

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            "x > 3: x was 0", tests.error.wo_mandatory_location(str(violation_err))
        )

    def test_condition_on_multiple_lines(self) -> None:
        # fmt: off
        @icontract.require(
            lambda x:
            x
            >
            3)
        def func(x: int) -> int:
            return x

        # fmt: on

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            textwrap.dedent(
                """\
                x
                    >
                    3: x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_err)),
        )

    def test_with_multiple_lambdas_on_a_line(self) -> None:
        # pylint: disable=unnecessary-lambda
        # fmt: off
        @icontract.require(
            error=lambda x: ValueError("x > 0, but got: {}".format(x)), condition=lambda x: x > 0)
        @icontract.require(
            error=lambda x: ValueError("x < 100, but got: {}".format(x)), condition=lambda x: x < 100)
        def func(x: int) -> int:
            return x

        # fmt: on

        value_error = None  # type: Optional[ValueError]
        try:
            func(x=101)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual("x < 100, but got: 101", str(value_error))

        value_error = None
        try:
            func(x=-1)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual("x > 0, but got: -1", str(value_error))


SOME_GLOBAL_CONSTANT = 10


class TestRepr(unittest.TestCase):
    def test_repr(self) -> None:
        a_repr = reprlib.Repr()
        a_repr.maxlist = 3

        @icontract.require(lambda x: len(x) < 10, a_repr=a_repr)
        def some_func(x: List[int]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=list(range(10 * 1000)))
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                len(x) < 10:
                len(x) was 10000
                x was [0, 1, 2, ...]"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestClass(unittest.TestCase):
    def test_nested_attribute(self) -> None:
        class B:
            def __init__(self) -> None:
                self.x = 0

            def x_plus_z(self, z: int) -> int:
                return self.x + z

            def __repr__(self) -> str:
                return "B(x={})".format(self.x)

        class A:
            def __init__(self) -> None:
                self.b = B()

            @icontract.require(lambda self: self.b.x > 0)
            def some_func(self) -> None:
                pass

            def __repr__(self) -> str:
                return "A()"

        a = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.some_func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.b.x > 0:
                self was A()
                self.b was B(x=0)
                self.b.x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_nested_method(self) -> None:
        z = 10

        class C:
            def __init__(self, x: int) -> None:
                self._x = x

            def x(self) -> int:
                return self._x

            def __repr__(self) -> str:
                return "C(x={})".format(self._x)

        class B:
            def c(self, x: int) -> C:
                return C(x=x)

            def __repr__(self) -> str:
                return "B()"

        def gt_zero(value: int) -> bool:
            return value > 0

        class A:
            def __init__(self) -> None:
                self.b = B()

            @icontract.require(
                lambda self: pathlib.Path(str(gt_zero(self.b.c(x=0).x() + 12.2 * z)))
                is None
            )
            def some_func(self) -> None:
                pass

            def __repr__(self) -> str:
                return "A()"

        a = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.some_func()
        except icontract.ViolationError as err:
            violation_error = err

        # This dummy path is necessary to obtain the class name.
        dummy_path = pathlib.Path("/just/a/dummy/path")

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                pathlib.Path(str(gt_zero(self.b.c(x=0).x() + 12.2 * z)))
                    is None:
                gt_zero(self.b.c(x=0).x() + 12.2 * z) was True
                pathlib.Path(str(gt_zero(self.b.c(x=0).x() + 12.2 * z))) was {}('True')
                self was A()
                self.b was B()
                self.b.c(x=0) was C(x=0)
                self.b.c(x=0).x() was 0
                str(gt_zero(self.b.c(x=0).x() + 12.2 * z)) was 'True'
                z was 10"""
            ).format(dummy_path.__class__.__name__),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestClosures(unittest.TestCase):
    def test_closure(self) -> None:
        y = 4
        z = 5

        @icontract.require(lambda x: x < y + z)
        def some_func(x: int) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=100)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < y + z:
                x was 100
                y was 4
                z was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_global(self) -> None:
        @icontract.require(lambda x: x < SOME_GLOBAL_CONSTANT)
        def some_func(x: int) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=100)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < SOME_GLOBAL_CONSTANT:
                SOME_GLOBAL_CONSTANT was 10
                x was 100"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_closure_and_global(self) -> None:
        y = 4

        @icontract.require(lambda x: x < y + SOME_GLOBAL_CONSTANT)
        def some_func(x: int) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=100)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < y + SOME_GLOBAL_CONSTANT:
                SOME_GLOBAL_CONSTANT was 10
                x was 100
                y was 4"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestWithNumpyMock(unittest.TestCase):
    def test_that_mock_works(self) -> None:
        arr = tests.mock.NumpyArray(values=[-3, 3])

        value_err = None  # type: Optional[ValueError]
        try:
            not (arr > 0)  # pylint: disable=superfluous-parens,unneeded-not
        except ValueError as err:
            value_err = err

        self.assertIsNotNone(value_err)
        self.assertEqual(
            "The truth value of an array with more than one element is ambiguous.",
            str(value_err),
        )

    def test_that_single_comparator_works(self) -> None:
        @icontract.require(lambda arr: (arr > 0).all())
        def some_func(arr: tests.mock.NumpyArray) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(arr=tests.mock.NumpyArray(values=[-3, 3]))
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                (arr > 0).all():
                (arr > 0).all() was False
                arr was NumpyArray([-3, 3])"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_that_multiple_comparators_fail(self) -> None:
        """
        Test that multiple comparators in an expression will fail.

        Multiple comparisons are not implemented in numpy as of version <= 1.16.
        The following snippet exemplifies the problem:

        .. code-block:: python

            import numpy as np

            x = np.array([-3, 3])
            -100 < x < 100
            Traceback (most recent call last):
            ...
            ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()

        :return:
        """

        @icontract.require(lambda arr: (-3 < arr < 0).all())
        def some_func(arr: tests.mock.NumpyArray) -> None:
            pass

        value_err = None  # type: Optional[ValueError]
        try:
            some_func(arr=tests.mock.NumpyArray(values=[-10, -1]))
        except ValueError as err:
            value_err = err

        self.assertIsNotNone(value_err)
        self.assertEqual(
            "The truth value of an array with more than one element is ambiguous.",
            str(value_err),
        )


class TestNumpyArrays(unittest.TestCase):
    def test_arange_in_args(self) -> None:
        # This test case addresses ``visit_Call` in the Visitor.
        # See: https://github.com/Parquery/icontract/issues/229

        @icontract.require(lambda arr: len(arr) > 2)
        def some_func(arr: Any) -> None:
            return

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(arr=numpy.arange(2))
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                len(arr) > 2:
                arr was array([0, 1])
                len(arr) was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_arange_in_kwargs_values(self) -> None:
        # This test case addresses ``visit_Call` in the Visitor.
        # See: https://github.com/Parquery/icontract/issues/229

        def custom_len(arr: Any) -> int:
            return len(arr)

        @icontract.require(lambda arr: custom_len(arr=arr) > 2)
        def some_func(arr: Any) -> None:
            return

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(arr=numpy.arange(2))
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                custom_len(arr=arr) > 2:
                arr was array([0, 1])
                custom_len(arr=arr) was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestRecomputationFailure(unittest.TestCase):
    def test_that_the_error_is_informative(self) -> None:
        counter = 0

        def some_condition() -> bool:
            nonlocal counter
            if counter == 0:
                # The first time we return False, so that the pre-condition fails.
                counter += 1
                return False

            # Every next time we raise the exception, so that the re-computation fails.
            raise RuntimeError("Recomputation shall fail.")

        @icontract.require(lambda: some_condition())
        def some_func() -> None:
            pass

        runtime_error = None  # type: Optional[RuntimeError]
        try:
            some_func()
        except RuntimeError as err:
            runtime_error = err

        assert runtime_error is not None

        lines = [
            re.sub(
                r"^File (.*), line ([0-9]+) in (.*):$",
                "File <erased path>, line <erased line> in <erased function>:",
                line,
            )
            for line in str(runtime_error).splitlines()
        ]
        text = "\n".join(lines)

        self.assertEqual(
            textwrap.dedent(
                """\
            Failed to recompute the values of the contract condition:
            File <erased path>, line <erased line> in <erased function>:
            lambda: some_condition()"""
            ),
            text,
        )


class TestTracingAll(unittest.TestCase):
    def test_global_variable(self) -> None:
        # fmt: off
        @icontract.require(
            lambda lst:
            all(
                value > SOME_GLOBAL_CONSTANT
                for value in lst
            )
        )
        # fmt: on
        def func(lst: List[int]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(lst=[-1, -2])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)

        got = tests.error.wo_mandatory_location(str(violation_error))

        self.assertEqual(
            textwrap.dedent(
                """\
                all(
                        value > SOME_GLOBAL_CONSTANT
                        for value in lst
                    ):
                SOME_GLOBAL_CONSTANT was 10
                all(
                        value > SOME_GLOBAL_CONSTANT
                        for value in lst
                    ) was False, e.g., with
                  value = -1
                lst was [-1, -2]"""
            ),
            got,
        )

    def test_formatted_string(self) -> None:
        # fmt: off
        @icontract.require(
            lambda lst:
            all(
                f'{value}' == 'x'
                for value in lst
            )
        )
        # fmt: on
        def func(lst: List[str]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(lst=["y"])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)

        got = tests.error.wo_mandatory_location(str(violation_error))

        self.assertEqual(
            textwrap.dedent(
                """\
                all(
                        f'{value}' == 'x'
                        for value in lst
                    ):
                all(
                        f'{value}' == 'x'
                        for value in lst
                    ) was False, e.g., with
                  value = 'y'
                lst was ['y']"""
            ),
            got,
        )

    def test_two_fors_and_two_ifs(self) -> None:
        # fmt: off
        @icontract.require(
            lambda matrix:
            all(
                cell > SOME_GLOBAL_CONSTANT
                for i, row in enumerate(matrix)
                if i > 0
                for j, cell in enumerate(row)
                if i == j
            )
        )
        # fmt: on
        def func(matrix: List[List[int]]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(matrix=[[-1, -1], [-1, -1]])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)

        got = re.sub(
            r"<enumerate object at 0x[0-9A-Za-z]+>",
            "<enumerate object at 0x...>",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        self.assertEqual(
            textwrap.dedent(
                """\
            all(
                    cell > SOME_GLOBAL_CONSTANT
                    for i, row in enumerate(matrix)
                    if i > 0
                    for j, cell in enumerate(row)
                    if i == j
                ):
            SOME_GLOBAL_CONSTANT was 10
            all(
                    cell > SOME_GLOBAL_CONSTANT
                    for i, row in enumerate(matrix)
                    if i > 0
                    for j, cell in enumerate(row)
                    if i == j
                ) was False, e.g., with
              i = 1
              row = [-1, -1]
              j = 1
              cell = -1
            enumerate(matrix) was <enumerate object at 0x...>
            matrix was [[-1, -1], [-1, -1]]"""
            ),
            got,
        )

    def test_nested_all(self) -> None:
        # Nesting is not recursively followed by design. Only the outer-most all expression should be traced.

        # fmt: off
        @icontract.require(
            lambda lst_of_lsts:
            all(
                all(item > 0 for item in sublst)
                for sublst in lst_of_lsts
            )
        )
        # fmt: on
        def func(lst_of_lsts: List[List[int]]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(lst_of_lsts=[[-1, -1]])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                all(
                        all(item > 0 for item in sublst)
                        for sublst in lst_of_lsts
                    ):
                all(
                        all(item > 0 for item in sublst)
                        for sublst in lst_of_lsts
                    ) was False, e.g., with
                  sublst = [-1, -1]
                lst_of_lsts was [[-1, -1]]"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_property_of_an_object_represented(self) -> None:
        class Something:
            def __init__(self) -> None:
                self.some_property = 0

            def __repr__(self) -> str:
                return "Something()"

        # fmt: off
        @icontract.require(
            lambda something, lst:
            all(
                item > something.some_property
                for item in lst
            )
        )
        # fmt: on
        def func(something: Something, lst: List[int]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(something=Something(), lst=[-1])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)

        self.assertEqual(
            textwrap.dedent(
                """\
                all(
                        item > something.some_property
                        for item in lst
                    ):
                all(
                        item > something.some_property
                        for item in lst
                    ) was False, e.g., with
                  item = -1
                lst was [-1]
                something was Something()
                something.some_property was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_shadows_in_targets(self) -> None:
        # fmt: off
        @icontract.require(
            lambda lst_of_lsts:
            all(
                all(item > 0 for item in item)
                for item in lst_of_lsts
            )
        )
        # fmt: on
        def func(lst_of_lsts: List[List[int]]) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            func(lst_of_lsts=[[-1, -1]])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                all(
                        all(item > 0 for item in item)
                        for item in lst_of_lsts
                    ):
                all(
                        all(item > 0 for item in item)
                        for item in lst_of_lsts
                    ) was False, e.g., with
                  item = [-1, -1]
                lst_of_lsts was [[-1, -1]]"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=unnecessary-lambda
import unittest
from typing import List

import icontract


class TestPrecondition(unittest.TestCase):
    def test_ok(self) -> None:
        order = []  # type: List[str]

        @icontract.require(lambda: another_func())
        @icontract.require(lambda: yet_another_func())
        def some_func() -> bool:
            order.append(some_func.__name__)
            return True

        @icontract.require(lambda: some_func())
        @icontract.require(
            lambda: yet_yet_another_func()
        )  # pylint: disable=unnecessary-lambda
        def another_func() -> bool:
            order.append(another_func.__name__)
            return True

        def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        some_func()

        self.assertListEqual(
            [
                "yet_another_func",
                "yet_yet_another_func",
                "some_func",
                "another_func",
                "some_func",
            ],
            order,
        )

    def test_recover_after_exception(self) -> None:
        order = []  # type: List[str]
        some_func_should_raise = True

        class CustomError(Exception):
            pass

        @icontract.require(lambda: another_func())
        @icontract.require(lambda: yet_another_func())
        def some_func() -> bool:
            order.append(some_func.__name__)
            if some_func_should_raise:
                raise CustomError("some_func_should_raise")
            return True

        @icontract.require(lambda: some_func())
        @icontract.require(lambda: yet_yet_another_func())
        def another_func() -> bool:
            order.append(another_func.__name__)
            return True

        def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        try:
            some_func()
        except CustomError:
            pass

        self.assertListEqual(
            ["yet_another_func", "yet_yet_another_func", "some_func"], order
        )

        # Reset for the next experiment
        order = []
        some_func_should_raise = False

        some_func()

        self.assertListEqual(
            [
                "yet_another_func",
                "yet_yet_another_func",
                "some_func",
                "another_func",
                "some_func",
            ],
            order,
        )


class TestPostcondition(unittest.TestCase):
    def test_ok(self) -> None:
        order = []  # type: List[str]
        another_func_should_raise = True

        class CustomError(Exception):
            pass

        @icontract.ensure(lambda: another_func())
        @icontract.ensure(lambda: yet_another_func())
        def some_func() -> bool:
            order.append(some_func.__name__)
            return True

        @icontract.ensure(lambda: some_func())
        @icontract.ensure(lambda: yet_yet_another_func())
        def another_func() -> bool:
            order.append(another_func.__name__)
            if another_func_should_raise:
                raise CustomError("some_func_should_raise")

            return True

        def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        try:
            some_func()
        except CustomError:
            pass

        self.assertListEqual(["some_func", "yet_another_func", "another_func"], order)

        # Reset for the next experiments
        order = []
        another_func_should_raise = False

        some_func()

        self.assertListEqual(
            [
                "some_func",
                "yet_another_func",
                "another_func",
                "yet_yet_another_func",
                "some_func",
            ],
            order,
        )

    def test_recover_after_exception(self) -> None:
        order = []  # type: List[str]

        @icontract.ensure(lambda: another_func())
        @icontract.ensure(lambda: yet_another_func())
        def some_func() -> bool:
            order.append(some_func.__name__)
            return True

        @icontract.ensure(lambda: some_func())
        @icontract.ensure(lambda: yet_yet_another_func())
        def another_func() -> bool:
            order.append(another_func.__name__)
            return True

        def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        some_func()

        self.assertListEqual(
            [
                "some_func",
                "yet_another_func",
                "another_func",
                "yet_yet_another_func",
                "some_func",
            ],
            order,
        )


class TestInvariant(unittest.TestCase):
    def test_ok(self) -> None:
        order = []  # type: List[str]

        @icontract.invariant(lambda self: self.some_func())
        class SomeClass(icontract.DBC):
            def __init__(self) -> None:
                order.append("__init__")

            def some_func(self) -> bool:
                order.append("some_func")
                return True

            def another_func(self) -> bool:
                order.append("another_func")
                return True

        some_instance = SomeClass()
        self.assertListEqual(["__init__", "some_func"], order)

        # Reset for the next experiment
        order = []

        some_instance.another_func()
        self.assertListEqual(["some_func", "another_func", "some_func"], order)

    def test_recover_after_exception(self) -> None:
        order = []  # type: List[str]
        some_func_should_raise = False

        class CustomError(Exception):
            pass

        @icontract.invariant(lambda self: self.some_func())
        class SomeClass(icontract.DBC):
            def __init__(self) -> None:
                order.append("__init__")

            def some_func(self) -> bool:
                order.append("some_func")
                if some_func_should_raise:
                    raise CustomError("some_func_should_raise")

                return True

            def another_func(self) -> bool:
                order.append("another_func")
                return True

        some_instance = SomeClass()
        self.assertListEqual(["__init__", "some_func"], order)

        # Reset for the next experiment
        order = []
        some_func_should_raise = True

        try:
            some_instance.another_func()
        except CustomError:
            pass

        self.assertListEqual(["some_func"], order)

        # Reset for the next experiment
        order = []
        some_func_should_raise = False

        some_instance.another_func()
        self.assertListEqual(["some_func", "another_func", "some_func"], order)

    def test_member_function_call_in_constructor(self) -> None:
        order = []  # type: List[str]

        @icontract.invariant(lambda self: self.some_attribute > 0)
        class SomeClass(icontract.DBC):
            def __init__(self) -> None:
                order.append("__init__ enters")
                self.some_attribute = self.some_func()
                order.append("__init__ exits")

            def some_func(self) -> int:
                order.append("some_func")
                return 3

        _ = SomeClass()
        self.assertListEqual(["__init__ enters", "some_func", "__init__ exits"], order)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name

import abc
import sys
import textwrap
import unittest
from typing import Optional, Sequence, cast  # pylint: disable=unused-import

import icontract
import tests.error


class TestOK(unittest.TestCase):
    def test_require_else(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x % 2 == 0)
            def func(self, x: int) -> None:
                pass

        class B(A):
            @icontract.require(lambda x: x % 3 == 0)
            def func(self, x: int) -> None:
                pass

        b = B()
        b.func(x=4)
        b.func(x=9)

    def test_triple_inheritance_with_require_else(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x % 2 == 0)
            def func(self, x: int) -> None:
                pass

        class B(A):
            @icontract.require(lambda x: x % 3 == 0)
            def func(self, x: int) -> None:
                pass

        class C(B):
            @icontract.require(lambda x: x % 5 == 0)
            def func(self, x: int) -> None:
                pass

        c = C()
        c.func(x=5)


class TestViolation(unittest.TestCase):
    def test_inherited_without_implementation(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x < 100)
            def func(self, x: int) -> None:
                pass

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func(x=1000)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < 100:
                self was an instance of B
                x was 1000"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_with_implementation(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x < 100)
            def func(self, x: int) -> None:
                pass

        class B(A):
            def func(self, x: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func(x=1000)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < 100:
                self was an instance of B
                x was 1000"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_require_else(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x % 2 == 0)
            def func(self, x: int) -> None:
                pass

        class B(A):
            @icontract.require(lambda x: x % 3 == 0)
            def func(self, x: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func(x=5)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x % 3 == 0:
                self was an instance of B
                x was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_triple_inheritance_wo_implementation(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x < 100)
            def func(self, x: int) -> None:
                pass

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        class C(B):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        c = C()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            c.func(x=1000)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < 100:
                self was an instance of C
                x was 1000"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_triple_inheritance_with_implementation(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x < 100)
            def func(self, x: int) -> None:
                pass

        class B(A):
            pass

        class C(B):
            def func(self, x: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        c = C()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            c.func(x=1000)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < 100:
                self was an instance of C
                x was 1000"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_triple_inheritance_with_require_else(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x % 2 == 0)
            def func(self, x: int) -> None:
                pass

        class B(A):
            @icontract.require(lambda x: x % 3 == 0)
            def func(self, x: int) -> None:
                pass

        class C(B):
            @icontract.require(lambda x: x % 5 == 0)
            def func(self, x: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        c = C()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            c.func(x=7)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x % 5 == 0:
                self was an instance of C
                x was 7"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_abstract_method(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            @abc.abstractmethod
            def func(self, x: int) -> int:
                pass

        class B(A):
            def func(self, x: int) -> int:
                return 1000

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > 0:
                self was an instance of B
                x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_that_base_preconditions_apply_to_init_if_not_defined(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x >= 0)
            def __init__(self, x: int) -> None:
                pass

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # Optional[icontract.ViolationError]
        try:
            _ = B(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x >= 0:
                self was an instance of B
                x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_that_base_preconditions_dont_apply_to_init_if_overridden(self) -> None:
        class A(icontract.DBC):
            @icontract.require(lambda x: x >= 0)
            def __init__(self, x: int) -> None:
                pass

        class B(A):
            # pylint: disable=super-init-not-called
            @icontract.require(lambda x: x < 0)
            def __init__(self, x: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        # Preconditions of B need to be satisfied, but not from A
        _ = B(x=-100)

        violation_error = None  # Optional[icontract.ViolationError]
        try:
            _ = B(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x < 0:
                self was an instance of B
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestPropertyOK(unittest.TestCase):
    def test_getter_setter_deleter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.deleted = False
                self._some_prop = 1

            @property
            @icontract.require(lambda self: self._some_prop > 0)
            def some_prop(self) -> int:
                return self._some_prop

            @some_prop.setter
            @icontract.require(lambda value: value > 0)
            def some_prop(self, value: int) -> None:
                self._some_prop = value

            @some_prop.deleter
            @icontract.require(lambda self: not self.deleted)
            def some_prop(self) -> None:
                self.deleted = True

        class SomeClass(SomeBase):
            pass

        some_inst = SomeClass()
        some_inst.some_prop = 3
        self.assertEqual(3, some_inst.some_prop)

        del some_inst.some_prop
        self.assertTrue(some_inst.deleted)


class TestPropertyViolation(unittest.TestCase):
    def test_getter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = True

            @property
            @icontract.require(lambda self: not self.toggled)
            def some_prop(self) -> int:
                return 0

        class SomeClass(SomeBase):
            @property
            def some_prop(self) -> int:
                return 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_setter(self) -> None:
        class SomeBase(icontract.DBC):
            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            @icontract.require(lambda value: value > 0)
            def some_prop(self, value: int) -> None:
                pass

        class SomeClass(SomeBase):
            @SomeBase.some_prop.setter  # type: ignore
            def some_prop(self, value: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = 0
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                value > 0:
                self was an instance of SomeClass
                value was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_deleter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = True

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.deleter
            @icontract.require(lambda self: not self.toggled)
            def some_prop(self) -> None:
                pass

        class SomeClass(SomeBase):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

            @SomeBase.some_prop.deleter  # type: ignore
            def some_prop(self) -> None:
                pass

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestConstructor(unittest.TestCase):
    def test_init_tightens_preconditions(self) -> None:
        class A(icontract.DBC):
            def __init__(self, x: int) -> None:
                pass

        class B(A):
            # B can require tighter pre-conditions than A.
            # __init__ is a special case: while other functions need to satisfy Liskov substitution principle,
            # __init__ is an exception.
            @icontract.require(lambda x: x > 0)
            def __init__(self, x: int) -> None:
                super().__init__(x=x)

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        _ = B(3)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > 0:
                self was an instance of B
                x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_new_tightens_preconditions(self) -> None:
        class A(icontract.DBC):
            def __new__(cls, xs: Sequence[int]) -> "A":
                return cast(A, xs)

        class B(A):
            # B can require tighter pre-conditions than A.
            # __new__ is a special case: while other functions need to satisfy Liskov substitution principle,
            # __new__ is an exception.
            @icontract.require(lambda xs: all(x > 0 for x in xs))
            def __new__(cls, xs: Sequence[int]) -> "B":
                return cast(B, xs)

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        _ = B([1, 2, 3])

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B([-1, -2, -3])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                all(x > 0 for x in xs):
                all(x > 0 for x in xs) was False, e.g., with
                  x = -1
                xs was [-1, -2, -3]"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestInvalid(unittest.TestCase):
    def test_abstract_method_not_implemented(self) -> None:
        # pylint: disable=abstract-method
        class A(icontract.DBC):
            @icontract.require(lambda x: x > 0)
            @abc.abstractmethod
            def func(self, x: int) -> int:
                pass

        class B(A):
            pass

        type_err = None  # type: Optional[TypeError]
        try:
            _ = B()  # type: ignore
        except TypeError as err:
            type_err = err

        self.assertIsNotNone(type_err)
        if sys.version_info < (3, 9):
            self.assertEqual(
                "Can't instantiate abstract class B with abstract methods func",
                str(type_err),
            )
        elif sys.version_info < (3, 12):
            self.assertEqual(
                "Can't instantiate abstract class B with abstract method func",
                str(type_err),
            )
        else:
            self.assertEqual(
                "Can't instantiate abstract class B without an implementation for abstract method 'func'",
                str(type_err),
            )

    def test_cant_weaken_base_function_without_preconditions(self) -> None:
        class A(icontract.DBC):
            @abc.abstractmethod
            def func(self, x: int) -> int:
                raise NotImplementedError()

        type_error = None  # type: Optional[TypeError]
        try:

            class B(A):  # pylint: disable=unused-variable
                @icontract.require(lambda x: x < 0)
                def func(self, x: int) -> int:
                    return 1000

        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertEqual(
            "The function "
            "TestInvalid.test_cant_weaken_base_function_without_preconditions.<locals>.B.func can not "
            "weaken the preconditions because the bases specify no preconditions at all. Hence this function must "
            "accept all possible input since the preconditions are OR'ed and no precondition implies a dummy "
            "precondition which is always fulfilled.",
            str(type_error),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring

import unittest

import icontract


class TestSlow(unittest.TestCase):
    def test_slow_set(self) -> None:
        self.assertTrue(
            icontract.SLOW,
            "icontract.SLOW was not set. Please check if you set the environment variable ICONTRACT_SLOW "
            "before running this test.",
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=unnecessary-lambda

import abc
import functools
import textwrap
import unittest
from typing import Optional, List, Type  # pylint: disable=unused-import

import icontract
from icontract._globals import CallableT
import tests.error
import tests.mock


class TestOK(unittest.TestCase):
    def test_with_condition_as_lambda(self) -> None:
        @icontract.ensure(lambda result, x: result > x)
        def some_func(x: int, y: int = 5) -> int:
            return x + y

        some_func(x=5)
        some_func(x=5, y=10)


class TestViolation(unittest.TestCase):
    def test_with_condition(self) -> None:
        @icontract.ensure(lambda result, x: result > x)
        def some_func(x: int, y: int = 5) -> int:
            return x - y

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result > x:
                result was -4
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function(self) -> None:
        def some_condition(result: int) -> bool:
            return result > 3

        @icontract.ensure(some_condition)
        def some_func(x: int) -> int:
            return x

        # Valid call
        some_func(x=4)

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                some_condition:
                result was 1
                x was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function_with_default_argument_value(self) -> None:
        def some_condition(result: int, y: int = 0) -> bool:
            return result > y

        @icontract.ensure(some_condition)
        def some_func(x: int) -> int:
            return x

        # Valid call
        some_func(x=1)

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                some_condition:
                result was -1
                x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function_with_default_argument_value_set(self) -> None:
        def some_condition(result: int, y: int = 0) -> bool:
            return result > y

        @icontract.ensure(some_condition)
        def some_func(x: int, y: int) -> int:
            return x

        # Valid call
        some_func(x=-1, y=-2)

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1, y=3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                some_condition:
                result was 1
                x was 1
                y was 3"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_description(self) -> None:
        @icontract.ensure(lambda result, x: result > x, "expected summation")
        def some_func(x: int, y: int = 5) -> int:
            return x - y

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                expected summation: result > x:
                result was -4
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_stacked_decorators(self) -> None:
        def mydecorator(f: CallableT) -> CallableT:
            @functools.wraps(f)
            def wrapped(*args, **kwargs):  # type: ignore
                result = f(*args, **kwargs)
                return result

            return wrapped  # type: ignore

        some_var = 1
        another_var = 2

        @mydecorator
        @icontract.ensure(lambda result, x: x < result + some_var)
        @icontract.ensure(lambda result, y: y > result + another_var)
        @mydecorator
        def some_func(x: int, y: int) -> int:
            return 100

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=0, y=10)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                y > result + another_var:
                another_var was 2
                result was 100
                x was 0
                y was 10"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_default_values_outer(self) -> None:
        @icontract.ensure(lambda result, c: result % c == 0)
        @icontract.ensure(lambda result, b: result < b)
        def some_func(a: int, b: int = 21, c: int = 2) -> int:
            return a

        # Check first the outer post condition
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(a=13)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result % c == 0:
                a was 13
                b was 21
                c was 2
                result was 13"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        # Check the inner post condition
        violation_error = None
        try:
            some_func(a=36)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result < b:
                a was 36
                b was 21
                c was 2
                result was 36"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_only_result(self) -> None:
        @icontract.ensure(lambda result: result > 3)
        def some_func(x: int) -> int:
            return 0

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=10 * 1000)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result > 3:
                result was 0
                x was 10000"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestError(unittest.TestCase):
    def test_as_type(self) -> None:
        @icontract.ensure(lambda result: result > 0, error=ValueError)
        def some_func(x: int) -> int:
            return x

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual(
            textwrap.dedent(
                """\
                result > 0:
                result was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(value_error)),
        )

    def test_as_function(self) -> None:
        @icontract.ensure(
            lambda result: result > 0,
            error=lambda result: ValueError("result must be positive."),
        )
        def some_func(x: int) -> int:
            return x

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("result must be positive.", str(value_error))

    def test_with_empty_args(self) -> None:
        @icontract.ensure(
            lambda result: result > 0,
            error=lambda: ValueError("result must be positive"),
        )
        def some_func(x: int) -> int:
            return x

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("result must be positive", str(value_error))

    def test_with_different_args_from_condition(self) -> None:
        @icontract.ensure(
            lambda result: result > 0,
            error=lambda x, result: ValueError(
                "x is {}, result is {}".format(x, result)
            ),
        )
        def some_func(x: int) -> int:
            return x

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x is 0, result is 0", str(value_error))


class TestToggling(unittest.TestCase):
    def test_disabled(self) -> None:
        @icontract.ensure(lambda x, result: x > result, enabled=False)
        def some_func(x: int) -> int:
            return 123

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            result = some_func(x=1234)
            self.assertEqual(123, result)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNone(violation_error)


class TestInClass(unittest.TestCase):
    def test_postcondition_in_static_method(self) -> None:
        class SomeClass:
            @staticmethod
            @icontract.ensure(lambda result: result != 0)
            def some_func(x: int) -> int:
                return x

        result = SomeClass.some_func(x=1)
        self.assertEqual(1, result)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass.some_func(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result != 0:
                result was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_postcondition_in_class_method(self) -> None:
        class SomeClass:
            @classmethod
            @icontract.ensure(lambda result: result != 0)
            def some_func(cls: Type["SomeClass"], x: int) -> int:
                return x

        result = SomeClass.some_func(x=1)
        self.assertEqual(1, result)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass.some_func(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result != 0:
                result was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_postcondition_in_abstract_static_method(self) -> None:
        class SomeAbstract(icontract.DBC):
            @staticmethod
            @abc.abstractmethod
            @icontract.ensure(lambda result: result != 0)
            def some_func(x: int) -> int:
                raise NotImplementedError()

        class SomeClass(SomeAbstract):
            @staticmethod
            def some_func(x: int) -> int:
                return x

        result = SomeClass.some_func(x=1)
        self.assertEqual(1, result)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass.some_func(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result != 0:
                result was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_postcondition_in_abstract_class_method(self) -> None:
        class Abstract(icontract.DBC):
            @classmethod
            @abc.abstractmethod
            @icontract.ensure(lambda result: result != 0)
            def some_func(cls: Type["Abstract"], x: int) -> int:
                pass

        class SomeClass(Abstract):
            @classmethod
            def some_func(cls: Type["SomeClass"], x: int) -> int:
                return x

        result = SomeClass.some_func(x=1)
        self.assertEqual(1, result)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass.some_func(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result != 0:
                result was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_getter(self) -> None:
        class SomeClass:
            def __init__(self) -> None:
                self._some_prop = -1

            @property
            @icontract.ensure(lambda result: result > 0)
            def some_prop(self) -> int:
                return self._some_prop

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result > 0:
                result was -1
                self was an instance of SomeClass"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_setter(self) -> None:
        class SomeClass:
            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            @icontract.ensure(lambda self: self.some_prop > 0)
            def some_prop(self, value: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = -1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.some_prop > 0:
                result was None
                self was an instance of SomeClass
                self.some_prop was 0
                value was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_deleter(self) -> None:
        class SomeClass:
            def __init__(self) -> None:
                self._some_prop = -1

            @property
            def some_prop(self) -> int:
                return self._some_prop

            @some_prop.deleter
            @icontract.ensure(lambda self: self.some_prop > 0)
            def some_prop(self) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.some_prop > 0:
                result was None
                self was an instance of SomeClass
                self.some_prop was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unnecessary-lambda
# pylint: disable=unused-argument

import textwrap
import unittest
from typing import List, Optional  # pylint: disable=unused-import

import icontract
import tests.error


class TestOK(unittest.TestCase):
    def test_without_argument(self) -> None:
        z = [1]

        @icontract.snapshot(lambda: z[:], name="z")
        @icontract.ensure(lambda OLD, val: OLD.z + [val] == z)
        def some_func(val: int) -> None:
            z.append(val)

        some_func(2)

    def test_with_name_same_for_single_argument(self) -> None:
        @icontract.snapshot(lambda lst: lst[:])
        @icontract.ensure(lambda OLD, val, lst: OLD.lst + [val] == lst)
        def some_func(lst: List[int], val: int) -> None:
            lst.append(val)

        # Expected to pass
        some_func([1], 2)

    def test_with_custom_name_for_single_argument(self) -> None:
        @icontract.snapshot(lambda lst: len(lst), name="len_lst")
        @icontract.ensure(lambda OLD, val, lst: OLD.len_lst + 1 == len(lst))
        def some_func(lst: List[int], val: int) -> None:
            lst.append(val)

        # Expected to pass
        some_func([1], 2)

    def test_with_multiple_arguments(self) -> None:
        @icontract.snapshot(lambda lst_a, lst_b: set(lst_a).union(lst_b), name="union")
        @icontract.ensure(
            lambda OLD, lst_a, lst_b: set(lst_a).union(lst_b) == OLD.union
        )
        def some_func(lst_a: List[int], lst_b: List[int]) -> None:
            pass

        # Expected to pass
        some_func(lst_a=[1, 2], lst_b=[3, 4])


class TestViolation(unittest.TestCase):
    def test_with_name_same_as_argument(self) -> None:
        @icontract.snapshot(lambda lst: lst[:])
        @icontract.ensure(lambda OLD, val, lst: OLD.lst + [val] == lst)
        def some_func(lst: List[int], val: int) -> None:
            lst.append(val)
            lst.append(1984)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func([1], 2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                OLD.lst + [val] == lst:
                OLD was a bunch of OLD values
                OLD.lst was [1]
                lst was [1, 2, 1984]
                result was None
                val was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_custom_name(self) -> None:
        @icontract.snapshot(lambda lst: len(lst), name="len_lst")
        @icontract.ensure(lambda OLD, val, lst: OLD.len_lst + 1 == len(lst))
        def some_func(lst: List[int], val: int) -> None:
            lst.append(val)
            lst.append(1984)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func([1], 2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                OLD.len_lst + 1 == len(lst):
                OLD was a bunch of OLD values
                OLD.len_lst was 1
                len(lst) was 3
                lst was [1, 2, 1984]
                result was None
                val was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_multiple_arguments(self) -> None:
        @icontract.snapshot(lambda lst_a, lst_b: set(lst_a).union(lst_b), name="union")
        @icontract.ensure(
            lambda OLD, lst_a, lst_b: set(lst_a).union(lst_b) == OLD.union
        )
        def some_func(lst_a: List[int], lst_b: List[int]) -> None:
            lst_a.append(1984)  # bug

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(lst_a=[1, 2], lst_b=[3, 4])
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                set(lst_a).union(lst_b) == OLD.union:
                OLD was a bunch of OLD values
                OLD.union was {1, 2, 3, 4}
                lst_a was [1, 2, 1984]
                lst_b was [3, 4]
                result was None
                set(lst_a) was {1, 2, 1984}
                set(lst_a).union(lst_b) was {1, 2, 3, 4, 1984}"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestInvalid(unittest.TestCase):
    def test_missing_snapshot_but_old_in_postcondition(self) -> None:
        @icontract.ensure(lambda OLD, val, lst: OLD.len_lst + 1 == len(lst))
        def some_func(lst: List[int], val: int) -> None:
            lst.append(val)

        type_error = None  # type: Optional[TypeError]
        try:
            some_func([1], 2)
        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertEqual(
            "The argument(s) of the contract condition have not been set: ['OLD']. "
            "Does the original function define them? Did you supply them in the call? "
            "Did you decorate the function with a snapshot to capture OLD values?",
            tests.error.wo_mandatory_location(str(type_error)),
        )

    def test_conflicting_snapshots_with_argument_name(self) -> None:
        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            @icontract.snapshot(lambda lst: lst[:])
            @icontract.snapshot(lambda lst: lst[:])
            @icontract.ensure(lambda OLD, val, lst: len(OLD.lst) + 1 == len(lst))
            def some_func(lst: List[int], val: int) -> None:
                lst.append(val)

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "There are conflicting snapshots with the name: 'lst'", str(value_error)
        )

    def test_conflicting_snapshots_with_custom_name(self) -> None:
        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            @icontract.snapshot(lambda lst: len(lst), name="len_lst")
            @icontract.snapshot(lambda lst: len(lst), name="len_lst")
            @icontract.ensure(lambda OLD, val, lst: OLD.len_lst + 1 == len(lst))
            def some_func(lst: List[int], val: int) -> None:
                lst.append(val)

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "There are conflicting snapshots with the name: 'len_lst'", str(value_error)
        )

    def test_with_invalid_argument(self) -> None:
        # lst versus a_list
        type_error = None  # type: Optional[TypeError]
        try:

            @icontract.snapshot(lambda lst: len(lst), name="len_lst")
            @icontract.ensure(lambda OLD, val, a_list: OLD.len_lst + 1 == len(a_list))
            def some_func(a_list: List[int], val: int) -> None:
                a_list.append(val)

            some_func([1], 2)
        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertEqual(
            "The argument(s) of the snapshot have not been set: ['lst']. "
            "Does the original function define them? Did you supply them in the call?",
            tests.error.wo_mandatory_location(str(type_error)),
        )

    def test_with_no_arguments_and_no_name(self) -> None:
        z = [1]

        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            @icontract.snapshot(lambda: z[:])
            @icontract.ensure(lambda OLD, val: OLD.z + [val] == z)
            def some_func(val: int) -> None:
                z.append(val)

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "You must name a snapshot if no argument was given in the capture function.",
            str(value_error),
        )

    def test_with_multiple_arguments_and_no_name(self) -> None:
        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            @icontract.snapshot(lambda lst_a, lst_b: set(lst_a).union(lst_b))
            @icontract.ensure(
                lambda OLD, lst_a, lst_b: set(lst_a).union(lst_b) == OLD.union
            )
            def some_func(lst_a: List[int], lst_b: List[int]) -> None:
                pass

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "You must name a snapshot if multiple arguments were given in the capture function.",
            str(value_error),
        )

    def test_with_no_postcondition(self) -> None:
        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            @icontract.snapshot(lambda lst: lst[:])
            def some_func(lst: List[int]) -> None:
                return

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "You are decorating a function with a snapshot, "
            "but no postcondition was defined on the function before.",
            str(value_error),
        )

    def test_missing_old_attribute(self) -> None:
        @icontract.snapshot(lambda lst: lst[:])
        @icontract.ensure(
            lambda OLD, lst: OLD.len_list == lst
        )  # We miss len_lst in OLD here!
        def some_func(lst: List[int]) -> None:
            return

        attribute_error = None  # type: Optional[AttributeError]

        try:
            some_func(lst=[1, 2, 3])
        except AttributeError as error:
            attribute_error = error

        assert attribute_error is not None

        self.assertEqual(
            "The snapshot with the name 'len_list' is not available in the OLD of a postcondition. "
            "Have you decorated the function with a corresponding snapshot decorator?",
            str(attribute_error),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=unnecessary-lambda

import functools
import pathlib
import textwrap
import time
import unittest
from typing import Optional  # pylint: disable=unused-import

import icontract
from icontract._globals import CallableT
import tests.error
import tests.mock


class TestOK(unittest.TestCase):
    def test_that_it_works(self) -> None:
        @icontract.require(lambda x: x > 3)
        def some_func(x: int, y: int = 5) -> None:
            pass

        some_func(x=5)
        some_func(x=5, y=10)


class TestViolation(unittest.TestCase):
    def test_only_with_condition_arg(self) -> None:
        @icontract.require(lambda x: x > 3)
        def some_func(x: int, y: int = 5) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > 3:
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_description(self) -> None:
        @icontract.require(lambda x: x > 3, "x must not be small")
        def some_func(x: int, y: int = 5) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x must not be small: x > 3:
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function(self) -> None:
        def some_condition(x: int) -> bool:
            return x > 3

        @icontract.require(some_condition)
        def some_func(x: int, y: int = 5) -> str:
            return str(x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                some_condition:
                x was 1
                y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function_with_default_argument_value(self) -> None:
        def some_condition(x: int, y: int = 0) -> bool:
            return x > y

        @icontract.require(some_condition)
        def some_func(x: int) -> None:
            pass

        # Valid call
        some_func(x=1)

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "some_condition: x was -1",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function_with_default_argument_value_set(self) -> None:
        def some_condition(x: int, y: int = 0) -> bool:
            return x > y

        @icontract.require(some_condition)
        def some_func(x: int, y: int) -> None:
            pass

        # Valid call
        some_func(x=3, y=1)

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=-1, y=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                some_condition:
                x was -1
                y was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_pathlib(self) -> None:
        @icontract.require(lambda path: path.exists())
        def some_func(path: pathlib.Path) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(path=pathlib.Path("/doesnt/exist/test_contract"))
        except icontract.ViolationError as err:
            violation_error = err

        # This dummy path is necessary to obtain the class name.
        dummy_path = pathlib.Path("/also/doesnt/exist")

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                path.exists():
                path was {}('/doesnt/exist/test_contract')
                path.exists() was False"""
            ).format(dummy_path.__class__.__name__),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_multiple_comparators(self) -> None:
        @icontract.require(lambda x: 0 < x < 3)
        def some_func(x: int) -> str:
            return str(x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=10)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "0 < x < 3: x was 10",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_stacked_decorators(self) -> None:
        def mydecorator(f: CallableT) -> CallableT:
            @functools.wraps(f)
            def wrapped(*args, **kwargs):  # type: ignore
                result = f(*args, **kwargs)
                return result

            return wrapped  # type: ignore

        some_var = 100
        another_var = 0

        @mydecorator
        @icontract.require(lambda x: x < some_var)
        @icontract.require(lambda x: x > another_var)
        @mydecorator
        def some_func(x: int) -> str:
            return str(x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > another_var:
                another_var was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_default_values(self) -> None:
        @icontract.require(lambda a: a < 10)
        @icontract.require(lambda b: b < 10)
        @icontract.require(lambda c: c < 10)
        def some_func(a: int, b: int = 21, c: int = 22) -> int:
            return a + b

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(a=2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                c < 10:
                a was 2
                b was 21
                c was 22"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        violation_error = None
        try:
            some_func(a=2, c=8)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                b < 10:
                a was 2
                b was 21
                c was 8"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestBenchmark(unittest.TestCase):
    @unittest.skip(
        "Skipped the benchmark, execute manually on a prepared benchmark machine."
    )
    def test_enabled(self) -> None:
        @icontract.require(lambda x: x > 3)
        def pow_with_pre(x: int, y: int) -> int:
            result = x**y
            assert isinstance(result, int)
            return result

        def pow_wo_pre(x: int, y: int) -> int:
            if x <= 3:
                raise ValueError("precondition")

            result = x**y
            assert isinstance(result, int)
            return result

        start = time.time()
        for i in range(5, 10 * 1000):
            pow_with_pre(x=i, y=2)
        duration_with_pre = time.time() - start

        start = time.time()
        for i in range(5, 10 * 1000):
            pow_wo_pre(x=i, y=2)
        duration_wo_pre = time.time() - start

        self.assertLess(duration_with_pre / duration_wo_pre, 6)

    @unittest.skip(
        "Skipped the benchmark, execute manually on a prepared benchmark machine."
    )
    def test_disabled(self) -> None:
        @icontract.require(lambda x: x > 3, enabled=False)
        def pow_with_pre(x: int, y: int) -> int:
            result = x**y
            assert isinstance(result, int)
            return result

        def pow_wo_pre(x: int, y: int) -> int:
            result = x**y
            assert isinstance(result, int)
            return result

        start = time.time()
        for i in range(5, 10 * 1000):
            pow_with_pre(x=i, y=2)
        duration_with_pre = time.time() - start

        start = time.time()
        for i in range(5, 10 * 1000):
            pow_wo_pre(x=i, y=2)
        duration_wo_pre = time.time() - start

        self.assertLess(duration_with_pre / duration_wo_pre, 1.2)


class TestError(unittest.TestCase):
    def test_as_type(self) -> None:
        @icontract.require(lambda x: x > 0, error=ValueError)
        def some_func(x: int) -> int:
            return 0

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual(
            "x > 0: x was 0", tests.error.wo_mandatory_location(str(value_error))
        )

    def test_as_function(self) -> None:
        @icontract.require(
            lambda x: x > 0, error=lambda x: ValueError("x non-negative")
        )
        def some_func(x: int) -> int:
            return 0

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x non-negative", str(value_error))

    def test_as_function_with_outer_scope(self) -> None:
        z = 42

        @icontract.require(
            lambda x: x > 0,
            error=lambda x: ValueError("x non-negative, z: {}".format(z)),
        )
        def some_func(x: int) -> int:
            return 0

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x non-negative, z: 42", str(value_error))

    def test_with_empty_args(self) -> None:
        @icontract.require(
            lambda x: x > 0, error=lambda: ValueError("x must be positive")
        )
        def some_func(x: int) -> int:
            return 0

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x must be positive", str(value_error))

    def test_with_different_args_from_condition(self) -> None:
        @icontract.require(
            lambda x: x > 0,
            error=lambda x, y: ValueError("x is {}, y is {}".format(x, y)),
        )
        def some_func(x: int, y: int) -> int:
            return 0

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=0, y=10)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x is 0, y is 10", str(value_error))


class TestToggling(unittest.TestCase):
    def test_enabled(self) -> None:
        @icontract.require(lambda x: x > 10, enabled=False)
        def some_func(x: int) -> int:
            return 123

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            result = some_func(x=0)
            self.assertEqual(123, result)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNone(violation_error)


class TestInClass(unittest.TestCase):
    def test_instance_method(self) -> None:
        class A:
            def __init__(self) -> None:
                self.y = 5

            @icontract.require(lambda x: x > 3)
            def some_method(self, x: int) -> int:
                return self.y

            @icontract.require(lambda self: self.y > 10)
            def some_method_with_self(self) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        a = A()

        # Test method without self
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.some_method(x=1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                x > 3:
                self was an instance of A
                x was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        # Test method with self
        violation_error = None
        try:
            a.some_method_with_self()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.y > 10:
                self was an instance of A
                self.y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_unbound_instance_method_with_self_as_kwarg(self) -> None:
        class A:
            def __init__(self) -> None:
                self.y = 5

            @icontract.require(lambda self: self.y > 10)
            def some_method_with_self(self) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        a = A()

        func = a.some_method_with_self.__func__  # type: ignore

        violation_error = None
        try:
            func(self=a)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.y > 10:
                self was an instance of A
                self.y was 5"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_getter(self) -> None:
        class SomeClass:
            def __init__(self) -> None:
                self._some_prop = -1

            @property
            @icontract.require(lambda self: self._some_prop > 0)
            def some_prop(self) -> int:
                return self._some_prop

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self._some_prop > 0:
                self was an instance of SomeClass
                self._some_prop was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_setter(self) -> None:
        class SomeClass:
            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            @icontract.require(lambda value: value > 0)
            def some_prop(self, value: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = -1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                value > 0:
                self was an instance of SomeClass
                value was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_deleter(self) -> None:
        class SomeClass:
            def __init__(self) -> None:
                self._some_prop = -1

            @property
            def some_prop(self) -> int:
                return self._some_prop

            @some_prop.deleter
            @icontract.require(lambda self: self.some_prop > 0)
            def some_prop(self) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.some_prop > 0:
                self was an instance of SomeClass
                self.some_prop was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestInvalid(unittest.TestCase):
    def test_unexpected_precondition_arguments(self) -> None:
        @icontract.require(lambda b: b > 3)
        def some_function(a: int) -> None:  # pylint: disable=unused-variable
            pass

        type_err = None  # type: Optional[TypeError]
        try:
            some_function(a=13)
        except TypeError as err:
            type_err = err

        self.assertIsNotNone(type_err)
        self.assertEqual(
            "The argument(s) of the contract condition have not been set: ['b']. "
            "Does the original function define them? Did you supply them in the call?",
            tests.error.wo_mandatory_location(str(type_err)),
        )

    def test_error_with_invalid_arguments(self) -> None:
        @icontract.require(
            lambda x: x > 0,
            error=lambda x, z: ValueError("x is {}, y is {}".format(x, z)),
        )
        def some_func(x: int, y: int) -> int:
            return 0

        type_error = None  # type: Optional[TypeError]
        try:
            some_func(x=0, y=10)
        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertEqual(
            "The argument(s) of the contract error have not been set: ['z']. "
            "Does the original function define them? Did you supply them in the call?",
            tests.error.wo_mandatory_location(str(type_error)),
        )

    def test_no_boolyness(self) -> None:
        @icontract.require(lambda: tests.mock.NumpyArray([True, False]))
        def some_func() -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "Failed to negate the evaluation of the condition.",
            tests.error.wo_mandatory_location(str(value_error)),
        )

    def test_unexpected_positional_argument(self) -> None:
        @icontract.require(lambda: True)
        def some_func() -> None:
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            some_func(0)  # type: ignore
        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertRegex(
            str(type_error),
            r"^([a-zA-Z_0-9<>.]+\.)?some_func\(\) takes 0 positional arguments but 1 was given$",
        )

    def test_unexpected_keyword_argument(self) -> None:
        @icontract.require(lambda: True)
        def some_func() -> None:
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            # pylint: disable=unexpected-keyword-arg
            some_func(x=0)  # type: ignore
        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertRegex(
            str(type_error),
            r"^([a-zA-Z_0-9<>.]+\.)?some_func\(\) got an unexpected keyword argument 'x'$",
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
import textwrap
import unittest
from typing import Optional, List  # pylint: disable=unused-import

import icontract

import tests.error


class TestOK(unittest.TestCase):
    def test_with_no_override(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.lst = []  # type: List[int]

            @icontract.snapshot(lambda self: self.lst[:], name="lst")
            @icontract.ensure(lambda OLD, self, val: OLD.lst + [val] == self.lst)
            def some_func(self, val: int) -> None:
                self.lst.append(val)

        class B(A):
            pass

        b = B()
        b.some_func(2)


class TestViolation(unittest.TestCase):
    def test_with_no_override(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.lst = []  # type: List[int]

            @icontract.snapshot(lambda self: self.lst[:], name="lst")
            @icontract.ensure(lambda OLD, self, val: OLD.lst + [val] == self.lst)
            def some_func(self, val: int) -> None:
                self.lst.append(val)
                self.lst.append(1984)

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.some_func(2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                OLD.lst + [val] == self.lst:
                OLD was a bunch of OLD values
                OLD.lst was []
                result was None
                self was an instance of B
                self.lst was [2, 1984]
                val was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_inherited_snapshot(self) -> None:
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.lst = []  # type: List[int]

            @icontract.snapshot(lambda self: len(self.lst), name="len_lst")
            @icontract.ensure(lambda self: self.lst)
            def some_func(self, val: int) -> None:
                pass

        class B(A):
            @icontract.ensure(lambda OLD, self: OLD.len_lst + 1 == len(self.lst))
            def some_func(self, val: int) -> None:
                self.lst.append(val)
                self.lst.append(1984)

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.some_func(2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                OLD.len_lst + 1 == len(self.lst):
                OLD was a bunch of OLD values
                OLD.len_lst was 0
                len(self.lst) was 2
                result was None
                self was an instance of B
                self.lst was [2, 1984]
                val was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestPropertyOK(unittest.TestCase):
    def test_getter_setter_deleter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.gets = 0
                self.sets = 0
                self.dels = 0

            @property
            @icontract.snapshot(lambda self: self.gets, name="gets")
            @icontract.ensure(lambda OLD, self: self.gets == OLD.gets + 1)
            def some_prop(self) -> int:
                self.gets += 1
                return 0

            @some_prop.setter
            @icontract.snapshot(lambda self: self.sets, name="sets")
            @icontract.ensure(lambda OLD, self: self.sets == OLD.sets + 1)
            def some_prop(self, value: int) -> None:
                # pylint: disable=unused-argument
                self.sets += 1

            @some_prop.deleter
            @icontract.snapshot(lambda self: self.dels, name="dels")
            @icontract.ensure(lambda OLD, self: self.dels == OLD.dels + 1)
            def some_prop(self) -> None:
                self.dels += 1

        class SomeClass(SomeBase):
            pass

        some_inst = SomeClass()
        _ = some_inst.some_prop
        some_inst.some_prop = 3
        del some_inst.some_prop

        self.assertEqual(1, some_inst.gets)
        self.assertEqual(1, some_inst.sets)
        self.assertEqual(1, some_inst.dels)


class TestPropertyViolation(unittest.TestCase):
    def test_getter_setter_deleter_fail(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.gets = 0
                self.sets = 0
                self.dels = 0

            @property
            @icontract.snapshot(lambda self: self.gets, name="gets")
            @icontract.ensure(lambda OLD, self: self.gets == OLD.gets + 1)
            def some_prop(self) -> int:
                # no self.gets increment
                return 0

            @some_prop.setter
            @icontract.snapshot(lambda self: self.sets, name="sets")
            @icontract.ensure(lambda OLD, self: self.sets == OLD.sets + 1)
            def some_prop(self, value: int) -> None:
                # pylint: disable=unused-argument
                # no self.sets increment
                return

            @some_prop.deleter
            @icontract.snapshot(lambda self: self.dels, name="dels")
            @icontract.ensure(lambda OLD, self: self.dels == OLD.dels + 1)
            def some_prop(self) -> None:
                # no self.dels increment
                return

        class SomeClass(SomeBase):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        # getter fails
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.gets == OLD.gets + 1:
                OLD was a bunch of OLD values
                OLD.gets was 0
                result was 0
                self was an instance of SomeClass
                self.gets was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        # setter fails
        violation_error = None
        try:
            some_inst.some_prop = 1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.sets == OLD.sets + 1:
                OLD was a bunch of OLD values
                OLD.sets was 0
                result was None
                self was an instance of SomeClass
                self.sets was 0
                value was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        # deleter fails
        violation_error = None
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.dels == OLD.dels + 1:
                OLD was a bunch of OLD values
                OLD.dels was 0
                result was None
                self was an instance of SomeClass
                self.dels was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestInvalid(unittest.TestCase):
    def test_conflicting_snapshot_names(self) -> None:
        value_error = None  # type: Optional[ValueError]
        try:

            class A(icontract.DBC):
                def __init__(self) -> None:
                    self.lst = []  # type: List[int]

                @icontract.snapshot(lambda self: len(self.lst), name="len_lst")
                @icontract.ensure(lambda self: self.lst)
                def some_func(self, val: int) -> None:
                    pass

            # pylint: disable=unused-variable

            class B(A):
                @icontract.snapshot(lambda self: len(self.lst), name="len_lst")
                @icontract.ensure(lambda OLD, self: OLD.len_lst + 1 == len(self.lst))
                def some_func(self, val: int) -> None:
                    self.lst.append(val)
                    self.lst.append(1984)

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "There are conflicting snapshots with the name: 'len_lst'.\n\n"
            "Please mind that the snapshots are inherited from the base classes. "
            "Does one of the base classes defines a snapshot with the same name?",
            str(value_error),
        )


class TestPropertyInvalid(unittest.TestCase):
    def test_getter_with_conflicting_snapshot_names(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.gets = 0

            @property
            @icontract.snapshot(lambda self: self.gets, name="gets")
            @icontract.ensure(lambda OLD, self: self.gets == OLD.gets + 1)
            def some_prop(self) -> int:
                self.gets += 1
                return 0

        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            class SomeClass(SomeBase):
                @property
                @icontract.snapshot(lambda self: self.gets, name="gets")
                @icontract.ensure(lambda OLD, self: self.gets == OLD.gets + 1)
                def some_prop(self) -> int:
                    return 0

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "There are conflicting snapshots with the name: 'gets'.\n\n"
            "Please mind that the snapshots are inherited from the base classes. "
            "Does one of the base classes defines a snapshot with the same name?",
            str(value_error),
        )

    def test_setter_with_conflicting_snapshot_names(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.sets = 0

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            @icontract.snapshot(lambda self: self.sets, name="sets")
            @icontract.ensure(lambda OLD, self: self.sets == OLD.sets + 1)
            def some_prop(self, value: int) -> None:
                # pylint: disable=unused-argument
                self.sets += 1

        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            class SomeClass(SomeBase):
                @property
                def some_prop(self) -> int:
                    return 0

                @some_prop.setter
                @icontract.snapshot(lambda self: self.sets, name="sets")
                @icontract.ensure(lambda OLD, self: self.sets == OLD.sets + 1)
                def some_prop(self, value: int) -> None:
                    # pylint: disable=unused-argument
                    return

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "There are conflicting snapshots with the name: 'sets'.\n\n"
            "Please mind that the snapshots are inherited from the base classes. "
            "Does one of the base classes defines a snapshot with the same name?",
            str(value_error),
        )

    def test_deleter_with_conflicting_snapshot_names(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.dels = 0

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.deleter
            @icontract.snapshot(lambda self: self.dels, name="dels")
            @icontract.ensure(lambda OLD, self: self.dels == OLD.dels + 1)
            def some_prop(self) -> None:
                self.dels += 1

        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable

            class SomeClass(SomeBase):
                @property
                def some_prop(self) -> int:
                    return 0

                @some_prop.deleter
                @icontract.snapshot(lambda self: self.dels, name="dels")
                @icontract.ensure(lambda OLD, self: self.dels == OLD.dels + 1)
                def some_prop(self) -> None:
                    return

        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "There are conflicting snapshots with the name: 'dels'.\n\n"
            "Please mind that the snapshots are inherited from the base classes. "
            "Does one of the base classes defines a snapshot with the same name?",
            str(value_error),
        )


if __name__ == "__main__":
    unittest.main()
